from collections import namedtuple
from dataclasses import dataclass
from typing import Any
from typing import Dict

from beancount.core.inventory import Inventory
from beanquery.query import run_query
from fava.application import render_template_string
from fava.context import g
from fava.core import FavaLedger
from fava.core.conversion import UNITS
from fava.core.query import COLUMNS
from fava.core.query import ObjectColumn
from fava.core.query import QueryResultTable
from fava.helpers import FavaAPIError


@dataclass(frozen=True)
class PanelCtx:
    ledger: Dict[str, Any]
    favaledger: FavaLedger
    panel: Dict[str, Any]


def render_template(ctx: PanelCtx, source: str) -> str:
    try:
        return render_template_string(
            source,
            # pass 'fava' for backwards compatibility
            fava=ctx.ledger,
            **ctx.__dict__,
        )
    except Exception as ex:
        raise FavaAPIError(f"failed to render template {source}: {ex}") from ex


def exec_query(ledger: FavaLedger, query):
    entries = g.filtered.entries_with_all_prices
    try:
        rtypes, rrows = run_query(entries, ledger.options, query)
    except Exception as ex:
        raise FavaAPIError(f'failed to execute query "{query}": {ex}') from ex

    querytable = _serialise(rtypes, rrows)

    # convert to legacy beancount.query format for backwards compat
    result_row = namedtuple("ResultRow", [col.name for col in rtypes])  # type: ignore[misc]
    rtypes = [(t.name, t.datatype) for t in rtypes]
    rrows = [result_row(*row) for row in rrows]

    return rtypes, rrows, querytable


def process_queries(ctx: PanelCtx):
    for query in ctx.panel.get("queries", []):
        if "bql" in query:
            bql = render_template(ctx, query["bql"])
            query["result_types"], query["result"], query["querytable"] = exec_query(ctx.favaledger, bql)


def process_jinja2(ctx: PanelCtx):
    if ctx.panel.get("type") != "jinja2":
        return

    template = ctx.panel.get("template", "")
    ctx.panel["template"] = render_template(ctx, template)


def sanitize_query_result(result):
    for i, row in enumerate(result):
        for k, v in row._asdict().items():
            if isinstance(v, Inventory):
                result[i] = result[i]._replace(**{k: UNITS.apply_inventory(v)})


def sanitize_panel(ctx):
    """replace or remove fields which are not JSON serializable"""
    for query in ctx.panel.get("queries", []):
        if "result" in query:
            sanitize_query_result(query["result"])

        if "result_types" in query:
            del query["result_types"]


def process_panel(ctx: PanelCtx):
    process_queries(ctx)
    process_jinja2(ctx)
    sanitize_panel(ctx)


# Copied from https://github.com/beancount/fava/blob/72d7504e6a86e72654d3974d2ca3ee3f3982f6ba/src/fava/core/query_shell.py#L242-L253
# Licensed under MIT License
def _serialise(rtypes, rrows) -> QueryResultTable:
    """Serialise the query result."""
    dtypes = [COLUMNS.get(c.datatype, ObjectColumn)(c.name) for c in rtypes]
    mappers = [d.serialise for d in dtypes]
    mapped_rows = [tuple(mapper(row[i]) for i, mapper in enumerate(mappers)) for row in rrows]
    return QueryResultTable(dtypes, mapped_rows)



#!/usr/bin/env python3
from __future__ import annotations

import argparse
import json
import re
import subprocess
import urllib.request
from datetime import datetime
from pathlib import Path

CHANGELOG_PATH = Path("CHANGELOG.md")
REPOSITORY = "andreasgerstmayr/fava-dashboards"


def run(cmd: list[str]) -> str:
    return subprocess.check_output(cmd, text=True).strip()


def last_tag() -> str:
    return run(["git", "describe", "--tags", "--abbrev=0"])


def prs_since(tag: str):
    output = run(["git", "log", f"{tag}..HEAD", "--pretty=format:%s"])
    for commit in output.splitlines():
        print(commit)
    return []


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("version", help="New version, for example v2.0.1")
    args = parser.parse_args()

    tag = run(["git", "describe", "--tags", "--abbrev=0"])
    commits = run(["git", "log", f"{tag}..HEAD", "--pretty=format:%s"]).splitlines()
    date = datetime.today().strftime("%Y-%m-%d")

    entries = []
    for commit in commits:
        m = re.search(r"\(#(\d+)\)$", commit)
        if not m:
            continue

        pr_id = m.group(1)
        print(f"Fetching PR #{pr_id} info...")
        with urllib.request.urlopen(f"https://api.github.com/repos/{REPOSITORY}/pulls/{pr_id}") as url:
            data = json.loads(url.read().decode())
            pr_title = data["title"]
            pr_author = data["user"]["login"]

        if pr_author in ["dependabot[bot]"]:
            continue

        entries.append(
            f"* {pr_title} [#{pr_id}](https://github.com/{REPOSITORY}/pull/{pr_id}) ([@{pr_author}](https://github.com/{pr_author}))"
        )

    entries.insert(0, f"## {args.version} ({date})")
    entries.append("")

    lines = CHANGELOG_PATH.read_text(encoding="utf8").splitlines()
    lines[2:2] = entries
    CHANGELOG_PATH.write_text("\n".join(lines) + "\n", encoding="utf8")

    print("Wrote changelog to", CHANGELOG_PATH)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import json
import re
import subprocess
from pathlib import Path

import yaml


def read_dashboards_utils(config_path: Path, config: dict):
    utils = config.get("utils", {})
    if "inline" in utils:
        return utils["inline"]
    elif "path" in utils:
        path = config_path.parent.joinpath(utils["path"])
        with open(path, "r", encoding="utf-8") as f:
            return f.read()
    else:
        return ""


def convert(in_path: Path, out_path: Path):
    with open(in_path, "r", encoding="utf8") as f:
        config = yaml.safe_load(f)

    if not config:
        config = {}

    utils = read_dashboards_utils(in_path, config).strip()
    if utils:
        utils = "const utils = (function(){" + utils + "})();"
        del config["utils"]

    print("Converting...")
    for dashboard in config.get("dashboards", []):
        for i, panel in enumerate(dashboard.get("panels", [])):
            if "type" in panel:
                panel_type = panel["type"]
                if panel_type not in ["html", "echarts", "d3_sankey"]:
                    title = panel.get("title", "?")
                    print(f'WARN: Skipping panel "{title}" with unsupported type "{panel_type}"')
                    del dashboard["panels"][i]
                    continue

                panel["kind"] = panel_type
                del panel["type"]

            if "script" in panel:
                script = panel["script"]
                if "queries" in panel:
                    queries = json.dumps(panel["queries"])
                    queries = queries.replace("{{ledger.ccy}}", '"+ledger.ccy+"')

                    script = (
                        "const panelQueries = await Promise.all("
                        + queries
                        + ".map(async (query) => ({...query, result: await ledger.query(query.bql)})));\n"
                        + script.replace("panel.queries", "panelQueries")
                    )
                    del panel["queries"]

                panel["spec"] = "async ({ ledger, variables }) => {\n" + script + "}"
                del panel["script"]

    # indent=2 required for multiline-regexp below
    config_json = json.dumps(config, indent=2, ensure_ascii=False)

    # convert spec to a function
    def add_fn(m):
        init = m.group(1)
        value = json.loads(m.group(2))
        return init + value

    config_tsx = re.sub(r'( *"spec": )(.+)$', add_fn, config_json, flags=re.MULTILINE)
    config_tsx = f"""/// <reference types="./fava-dashboards.d.ts" />
import {{ defineConfig }} from "fava-dashboards";

{utils}

export default defineConfig({config_tsx});
"""

    with open(out_path, "w", encoding="utf8") as f:
        f.write(config_tsx)

    print("Formatting...")
    subprocess.run(
        ["npx", "prettier", "-w", out_path.absolute()],
        cwd=Path(__file__).parent.parent.joinpath("frontend"),
        check=True,
        stdout=subprocess.DEVNULL,
    )

    print(f"File saved to {out_path}")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--output", default="dashboards.tsx", help="path to dashboards.tsx", type=Path)
    parser.add_argument("input", default="dashboards.yaml", help="path to dashboards.yaml", type=Path)
    args = parser.parse_args()

    convert(args.input, args.output)



import unittest
from datetime import date

from fava_dashboards.utils import clamp_to_ledger_range


class TestUtils(unittest.TestCase):
    def test_clamp(self):
        # filter contained in ledger
        assert clamp_to_ledger_range(date(2020, 1, 1), date(2021, 12, 31), date(2000, 1, 1), date(2024, 5, 1)) == (
            date(2020, 1, 1),
            date(2021, 12, 31),
        )

        # filter start outside of ledger
        assert clamp_to_ledger_range(date(1999, 1, 1), date(2024, 1, 1), date(2000, 1, 1), date(2024, 5, 1)) == (
            date(2000, 1, 1),
            date(2024, 1, 1),
        )

        # filter end outside of ledger
        assert clamp_to_ledger_range(date(2023, 1, 1), date(2024, 12, 31), date(2000, 1, 1), date(2024, 5, 1)) == (
            date(2023, 1, 1),
            date(2024, 5, 1),
        )

        # filter outside of ledger
        assert clamp_to_ledger_range(date(1990, 1, 1), date(1991, 12, 31), date(2000, 1, 1), date(2024, 5, 1)) == (
            date(1990, 1, 1),
            date(1991, 12, 31),
        )



import datetime
import functools
import json
import logging
import os
import traceback
from dataclasses import asdict
from dataclasses import dataclass
from pathlib import Path
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from fava.context import g
from fava.ext import FavaExtensionBase
from fava.ext import extension_endpoint
from fava.helpers import FavaAPIError
from flask import request

from fava_dashboards import legacy
from fava_dashboards.utils import clamp_to_ledger_range
from fava_dashboards.utils import get_ledger_duration
from fava_dashboards.utils import read_dashboards_tsx
from fava_dashboards.utils import read_dashboards_yaml

logger = logging.getLogger(__name__)
if loglevel := os.environ.get("LOGLEVEL"):
    logger.setLevel(loglevel.upper())


@dataclass(frozen=True)
class ExtConfig:
    dashboards_path: Path


@dataclass(frozen=True)
class LedgerData:  # pylint: disable=invalid-name,too-many-instance-attributes
    dateFirst: datetime.date
    dateLast: datetime.date
    filterFirst: Optional[datetime.date]
    filterLast: Optional[datetime.date]
    operatingCurrencies: List[str]
    ccy: str
    accounts: Dict[str, Any]
    commodities: Dict[str, Any]


def api_response(func):
    """return {success: true, data: ...} or {success: false, error: ...}"""

    @functools.wraps(func)
    def decorator(*args, **kwargs):
        try:
            data = func(*args, **kwargs)
            return {"success": True, "data": data}
        except FavaAPIError as e:
            return {"success": False, "error": e.message}, 500
        except Exception as e:  # pylint: disable=broad-exception-caught
            traceback.print_exception(e)
            return {"success": False, "error": str(e)}, 500

    return decorator


class FavaDashboards(FavaExtensionBase):
    report_title = "Dashboards"
    has_js_module = True

    def read_ext_config(self) -> ExtConfig:
        cfg = self.config if isinstance(self.config, dict) else {}

        if "config" in cfg:
            dashboards_path = self.ledger.join_path(cfg["config"])
        else:
            default_paths = [
                self.ledger.join_path(path)
                for path in ["dashboards.tsx", "dashboards.ts", "dashboards.jsx", "dashboards.js", "dashboards.yaml"]
            ]
            dashboards_path = default_paths[0]
            for path in default_paths:
                if path.exists():
                    dashboards_path = path
                    break

        return ExtConfig(dashboards_path=dashboards_path)

    def read_dashboards_utils(self, dashboards_yaml):
        utils = dashboards_yaml.get("utils", {})
        if "inline" in utils:
            return utils["inline"]
        elif "path" in utils:
            path = self.ledger.join_path(utils["path"])
            try:
                with open(path, "r", encoding="utf-8") as f:
                    return f.read()
            except Exception as ex:
                raise FavaAPIError(f"cannot read utils file {path}: {ex}") from ex
        else:
            return ""

    def get_ledger_data(self):
        operating_currencies = self.ledger.options["operating_currency"]
        if len(operating_currencies) == 0:
            raise FavaAPIError("no operating currency specified in the ledger")

        if g.filtered.date_range:
            filter_first = g.filtered.date_range.begin
            filter_last = g.filtered.date_range.end - datetime.timedelta(days=1)
            # Use filtered ledger here, as another filter (e.g. tag filter) could be applied.
            ledger_date_first, ledger_date_last = get_ledger_duration(g.filtered.entries)
            # Adjust the dates to align with ledger data.
            date_first, date_last = clamp_to_ledger_range(
                filter_first, filter_last, ledger_date_first, ledger_date_last
            )
        else:
            # No time filter applied.
            filter_first = None
            filter_last = None
            # Use filtered ledger here, as another filter (e.g. tag filter) could be applied.
            date_first, date_last = get_ledger_duration(g.filtered.entries)

        commodities = {c.currency: c for c in self.ledger.all_entries_by_type.Commodity}
        accounts = self.ledger.accounts
        return LedgerData(
            dateFirst=date_first,
            dateLast=date_last,
            filterFirst=filter_first,
            filterLast=filter_last,
            operatingCurrencies=list(operating_currencies),
            ccy=operating_currencies[0],
            accounts=accounts,
            commodities=commodities,
        )

    @extension_endpoint("query")
    @api_response
    def api_v1_query(self):
        bql = request.args.get("bql")

        _, result, _ = legacy.exec_query(self.ledger, bql)

        legacy.sanitize_query_result(result)
        return {"result": result}

    @extension_endpoint("v1_render_panel", ["POST"])
    @api_response
    def api_v1_render_panel(self):
        ledger = asdict(self.get_ledger_data())
        panel = request.get_json().get("panel", {})

        ctx = legacy.PanelCtx(ledger=ledger, favaledger=self.ledger, panel=panel)
        try:
            legacy.process_panel(ctx)
        except Exception as ex:
            raise FavaAPIError(f'error processing panel "{panel.get("title", "")}": {ex}') from ex

        return {"panel": ctx.panel}

    @extension_endpoint("v2_config")
    @api_response
    def api_v2_config(self):
        ext_config = self.read_ext_config()
        ledger_data = self.get_ledger_data()

        dashboard_format = ext_config.dashboards_path.suffix
        if dashboard_format == ".tsx":
            config_js = read_dashboards_tsx(ext_config.dashboards_path)
            return {"ledgerData": ledger_data, "configJs": config_js}
        elif dashboard_format == ".yaml":
            # backwards compat
            dashboards_yaml = read_dashboards_yaml(ext_config.dashboards_path)
            if dashboards_yaml is None:
                dashboards_yaml = {}
            config_js = "export default " + json.dumps(dashboards_yaml)
            utils = self.read_dashboards_utils(dashboards_yaml)
            return {"ledgerData": ledger_data, "configJs": config_js, "utilsJs": utils}
        else:
            raise FavaAPIError(f'invalid dashboard file "{ext_config.dashboards_path}"')



from datetime import date
from typing import List

import yaml
from fava.beans.abc import Directive
from fava.beans.abc import Price
from fava.beans.abc import Transaction
from fava.helpers import FavaAPIError


def read_dashboards_yaml(path: str):
    try:
        with open(path, "r", encoding="utf-8") as f:
            return yaml.safe_load(f)
    except Exception as ex:
        raise FavaAPIError(
            f"cannot read configuration file {path}: {ex}. To get started, use this template: https://github.com/andreasgerstmayr/fava-dashboards/blob/main/example/dashboards.tsx"
        ) from ex


def read_dashboards_tsx(path: str):
    try:
        with open(path, "r", encoding="utf-8") as f:
            return f.read()
    except Exception as ex:
        raise FavaAPIError(
            f"cannot read configuration file {path}: {ex}. To get started, use this template: https://github.com/andreasgerstmayr/fava-dashboards/blob/main/example/dashboards.tsx"
        ) from ex


def get_ledger_duration(entries: List[Directive]):
    date_first = None
    date_last = None
    for entry in entries:
        if isinstance(entry, Transaction):
            date_first = entry.date
            break
    for entry in reversed(entries):
        if isinstance(entry, (Transaction, Price)):
            date_last = entry.date
            break
    if not date_first or not date_last:
        raise FavaAPIError("no transaction found")
    return (date_first, date_last)


def clamp_to_ledger_range(filter_first: date, filter_last: date, ledger_first: date, ledger_last: date):
    # Adjust the dates in case the date filter is set to e.g. 2023-2024, however the ledger only contains data up to summer 2024.
    # Without this, all averages in the dashboard are off, because a wrong number of days between dateFirst and dateLast is calculated.
    clamped_first = max(filter_first, ledger_first)
    clamped_last = min(filter_last, ledger_last)

    # If there is no overlap of ledger and filter dates, leave them as-is.
    # For example filter: 2020-2021, but ledger data goes from 2022-2023.
    # The clamped date gives us from max(2020,2022) until min(2021,2023) = from 2022 until 2021, which is invalid.
    if clamped_first > clamped_last:
        return filter_first, filter_last

    return clamped_first, clamped_last
